//routines for handling roms

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <process.h>
#include <sys\stat.h>

#include "types.h"

#include "message.h"
#include "file.h"

#include "m6502\m6502.h"

#include "keyb.h"

#define FREE(x) if (x) {free(x); x=0;}
#define DELETE(x) if (x) {delete x; x=0;}

#include "nes.h"
#include "nesvideo.h"

#include "r2img.h"

//rom stuff
NESROMHEADER romhdr;
char *trainer;
int numrom,numvrom;
ROMPAGE *ROM;
VROMPAGE *VROM;

//
char *ram; //CPU address space
NES_ppumemory *ppu; //PPU address space
NES_sprite *spritemem; //Sprite memory

BATTERYMEM *batterymem; //battery backed memory

int NESinitialized=0;

int romloaded=0;
void freerom();

//rom filename (without extension) eg "CASTLEVA"
char romfilename[128];

//-----------------------------------------------------
// NES rom loader

void loadrom(char *romfile)
{
 if (!romfile) return;

 //free old rom
 freerom();

 //get rom name without extension
 strcpy(romfilename,romfile);
 if (strchr(romfilename,'.')) *strchr(romfilename,'.')=0;

 char s[64];
 strcpy(s,romfilename);
 strcat(s,".nes"); //add extension


 msg.printf(2,"Loading rom %s...",s);
 FILEIO f;
 if (f.open(s)) {msg.error("Unable to open file %s",s); return;}

  //read header
 if (f.read(&romhdr,sizeof(romhdr))) {msg.error("Unable to read header"); return;}

 if (!romhdr.validate()) {msg.error("Bad ROM header"); return;}

 //load trainer
 if (romhdr.trainer) trainer=(char *)f.readalloc(512);

 //read ROM
 numrom=romhdr.num16k;
 ROM=(ROMPAGE *)f.readalloc(numrom*sizeof(ROMPAGE));
 if (ROM) msg.printf(1,"%dK ROM read",numrom*sizeof(ROMPAGE)/1024);

 //read VROM
 numvrom=romhdr.num8k;
 VROM=(VROMPAGE *)f.readalloc(numvrom*sizeof(VROMPAGE)); //read vrom
 if (VROM) msg.printf(1,"%dK VROM read",numvrom*sizeof(VROMPAGE)/1024);

// msg.printf(1,"%d bytes left at end of .NES file",f.size()-f.getpos());

 f.close();
 msg.printf(2,"%s loaded",romfile);
 romloaded=1;

 if (romhdr.battery)
  {
   batterymem=new BATTERYMEM;
   batterymem->read(romfilename); //read battery backed rom
  } else batterymem=0;

 //initialize nes hardware....
 initNEShardware();

 //create video space
 new nesvideo(romfilename);
}

void freerom()
{
 if (batterymem)
  {
   batterymem->copyfromRAM();
   batterymem->write(romfilename); //write battery memory!
  }

 terminateNEShardware();
 FREE(trainer);
 FREE(ROM);
 FREE(VROM);
 FREE(batterymem);

 if (romloaded) msg.printf(2,"ROM freed");
 romloaded=0;

 //delete video space
 if (nv && nv->parent)  delete nv->parent;
}

void m_showmessages();
void m_getrominfo()
{
 if (!romloaded) msg.error("ROM not loaded");
  else  romhdr.printinfo(); //print info about rom
 m_showmessages();
}

void m_free()
{
 if (!romloaded) msg.error("ROM not loaded");
  else freerom();
}

void m_ramdump()
{
 if (!romloaded) msg.error("ROM not loaded");

 char s[64];

 for (int i=0; i<99; i++)
  {
   sprintf(s,"%s.d%02d",romfilename,i);
   struct stat sb;
   if (stat(s,&sb)!=0) break;
  }

 FILEIO f;
 f.create(s);
 f.write(ram,0x8000);
 f.write(m6502rom+0x8000,0x8000);
 f.close();

 msg.printf(1,"RAM dumped to %s",s);
}



void m_romdump()
{
 if (!romloaded) msg.error("ROM not loaded");
 char s[64];
 for (int i=0; i<numrom; i++)
  {
   sprintf(s,"%s.R%02X",romfilename,i);
   FILEIO f;
   f.create(s);
   f.write(ROM[i],sizeof(ROMPAGE));
   f.close();
  }
 msg.printf(1,"ROM dumped to %s.Rxx",romfilename);
}

//---------------------------------------------------------
//Battery backed memory

BATTERYMEM::BATTERYMEM()
{
  memset(data,0,0x2000);
}

void BATTERYMEM::read(char *romfilename)
{
 char s[64];
 strcpy(s,romfilename);
 strcat(s,".sav");

 //load savegame
 static FILEIO f;
 if (!f.open(s)) //try and open savegame...
   {
    f.read(data,0x2000); //read it!
    f.close();
    msg.printf(1,"Save game loaded from %s",s);
   }
}

void BATTERYMEM::write(char *romfilename)
{
 static char s[64];
 strcpy(s,romfilename);
 strcat(s,".sav");

 //save savegame
 static FILEIO f;
 f.create(s);
 f.write(data,0x2000); //read it!
 f.close();
 msg.printf(1,"Save game saved to %s",s);
}

void BATTERYMEM::copyfromRAM()
{
 if (!ram) return;
 memcpy(data,&ram[0x6000],0x2000);
}

void BATTERYMEM::copytoRAM()
{
 if (!ram) return;
 memcpy(&ram[0x6000],data,0x2000);
}



//---------------------------------------------------------
//NES rom header

char *getmmctypestr(int type);

int NESROMHEADER::validate() {return (str[0]=='N' && str[1]=='E' && str[2]=='S');}
void NESROMHEADER::printinfo()
{
 msg.printf(1,"# of 16K ROM banks: %d",num16k);
 msg.printf(1,"# of  8K VROM banks: %d",num8k);

 msg.printf(1,"%s mirroring",mirroring ? "Vertical" : "Horizontal");
 if (battery) msg.printf(1,"Battery backed RAM");
 if (trainer) msg.printf(1,"Trainer");
 if (fourscreen) msg.printf(1,"4 Screen VRAM");


 msg.printf(1,"MMC #%d: %s",banktype,getmmctypestr(banktype));
}


void m_writeromheader()
{
 if (!romloaded) msg.error("ROM not loaded");
 if (!nv) return;

 //put mirroring in rom
 romhdr.mirroring=nv->mirroring;

 char s[64];
 strcpy(s,romfilename);
 strcat(s,".nes");

 FILEIO f;
 if (f.open(s)) return;
 f.write(&romhdr,sizeof(romhdr));
 f.close();
 msg.printf(1,"NES ROM header written to %s",s);
}

extern COLOR nespal[256];

void m_savepalette()
{
 if (!romloaded) msg.error("ROM not loaded");
 if (!nv) return;
 FILEIO f;
 f.create("aNES.pal");
 f.write(nespal,256*sizeof(COLOR));
 f.close();
 msg.printf(1,"NES Palette saved to aNES.pal");
}



  /*
//----------------------------------------------------
//loading and saving game

void m_savegame()
{
 if (!romloaded) {msg.error("ROM not loaded"); return;}
}
*/













